from chem_analysis.analysis.logger import logger_analysis
from chem_analysis.analysis.base_obj.calibration import Cal
from chem_analysis.analysis.base_obj.chromatogram import Chromatogram
from chem_analysis.analysis.base_obj.signal_ import Signal

from chem_analysis.analysis.sec.sec_chrom import SECChrom
from chem_analysis.analysis.sec.sec_signal import SECSignal




import re

def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """ sorts in human order """
    return [atoi(c) for c in re.split(r'(\d+)', text)][1]


def get_files(path, pattern: str = "*.csv", sort_fun=natural_keys) -> list[str]:
    """ Returns a list of file names. """
    import glob
    import os

    # Find all json files
    file_list = []
    os.chdir(path)
    for files in glob.glob(pattern):
        file_list.append(files)  # filename with extension

    file_list.sort(key=sort_fun)
    return file_list



from dash import html, dcc

tab_process_layout = html.Div([
    html.H1('Page 2'),
    dcc.RadioItems(
        id='page-2-radios',
        options=[{'label': i, 'value': i} for i in ['Orange', 'Blue', 'Red']],
        value='Orange'
    ),
    html.Div(id='page-2-content')
])




import logging
import difflib


class ObjList:
    """ Object List
    Attributes
    ----------
    objs: list[Any]
        List of objs
    obj: Any
        type of object being stored
    limit: int
        maximum number of reference allowed
        default is 1000
    Methods
    -------
    add(item)
        Adds item to reference list
    remove(item)
        Removes item from reference list
    """

    def __init__(self, obj, add_objs=None, _logger: logging.Logger = None, limit: int = 1000):
        """
        Parameters
        ----------
        obj:
            the object the ref
        add_objs:
            objects will be passed to self.add()
        _logger:
            error class of parent node
        limit: int
            maximum number of reference allowed
            default is 10000
        """
        self._objs = []
        self._obj = obj
        self.limit = limit
        self.count = 0

        if _logger is None:
            self._logger = logging
        else:
            self._logger = _logger

        if add_objs is not None:
            self.add(add_objs)

    def __repr__(self):
        if self.count < 4:
            return "; ".join([repr(obj) for obj in self.objs])

        return "; ".join([repr(obj) for obj in self.objs[:2]]) + "; ..."

    def __call__(self):
        return self.objs

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.objs[item]
        elif isinstance(item, str):
            index = self._get_index_from_name(item)
            return self.objs[index]
        elif isinstance(item, slice):
            return [self.objs[i] for i in range(*item.indices(len(self.objs)))]
        else:
            mes = f"{item} not found."
            self._logger.error(mes)
            raise ValueError(mes)

    def __len__(self):
        return len(self._objs)

    def __iter__(self):
        for obj in self._objs:
            yield obj

    @property
    def objs(self):
        return self._objs

    def _get_index_from_name(self, item: str) -> int:
        """ get index from name
        Given an item name, return item index in list.
        This matching is done with difflib, so slight typing errors won't result in errors.
        Parameters
        ----------
        item: str
            name of item you are trying to find
        Returns
        -------
        index: int
            index of item in self._reference list
        Raises
        ------
        Exception
            If item name not found.
        """
        values = [i.key for i in self._objs]
        text, score = difflib.get_close_matches(item, values, n=1, cutoff=0.8)
        if score > 50:
            return values.index(text)
        else:
            mes = f"'{item}' not found."
            self._logger.error(mes)
            raise ValueError(mes)

    def add(self, objs):
        """ Add
        Adds object to reference list.
        Parameters
        ----------
        objs:
            object that you want to add to the list
        Raises
        -------
        Exception
            If invalid object is provided. An object that does not lead to a valid reference.
        """
        if not isinstance(objs, list):
            objs = [objs]

        add = []
        for obj in objs:
            if not isinstance(obj, self._obj):
                raise TypeError(f"expected: {self._obj}, received: {type(obj)} ")

            if objs in self._objs:  # check if reference s not already in list, and add it.
                self._logger.warning(f"'{obj}' already in list.")
                continue

            if hasattr(obj, "id_") and obj.id_ is None:
                obj.id_ = self.count + len(add)

            add.append(obj)

        self._objs += add
        self.count += len(add)

    def remove(self, objs):
        """ Remove
        Removes object from reference list.
        Parameters
        ----------
        objs:
            object that you want to remove to the list
        """
        if not isinstance(objs, list):
            objs = [objs]

        remove = []
        for obj in objs:
            if isinstance(obj, (str, int, slice)):
                obj = self[obj]

            if obj not in self._objs:
                self._logger.error(f"'{self._obj}'is not in list, so it can't be removed.")
                continue

            remove.append(obj)

        if not remove:
            return

        # loop through 'remove list' to remove objs
        for obj in remove:
            self._objs.remove(obj)
            self.count -= 1

    def clear(self):
        self._objs = []
        self.count = 0

    def as_dict(self) -> list:
        """ Returns list of references for serialization."""
        return [obj.as_dict() for obj in self.objs]




import pandas as pd
from dash.dependencies import Input, Output, State
from dash.dash_table.Format import Format, Scheme
from dash.exceptions import PreventUpdate

from chem_analysis.server import app
from chem_analysis.analysis.utils.load_csv import load_csv


@app.callback([Output('output-data-upload', 'children'), Output('dataframe', 'data')],
              [Input('upload-data', 'contents')],
              [State('upload-data', 'filename')])
def upload_file(list_of_contents, filename):
    if list_of_contents is not None:
        result, df = load_csv(list_of_contents, filename)
        if df is not None:
            print(df.head())
            df = df.to_json(orient='split')
        return result, df
    return None, {}


@app.callback([Output('data_table', 'data'), Output('data_table', 'columns')],
              [Input('dataframe', 'data')])
def set_table(json_data):
    if json_data:
        print("hi")
        df = pd.read_json(json_data, orient='split')
        data = df.iloc[:10, :].to_dict('records')
        columns = []
        for i in df.columns:
            columns.append({
                "name": i,
                "id": i,
                "renamable": True,
                "deletable": True,
                "type": 'numeric',
                "format": Format(precision=2, scheme=Scheme.exponent) #precision=2, scheme=Scheme.fixed  https://dash.plotly.com/datatable/data-formatting
            })

        return data, columns
    raise PreventUpdate


@app.callback(Output("ri_time", "options"),
              Input('dataframe', 'data'))
def update_radio(json_data):
    if json_data:
        df = pd.read_json(json_data, orient='split')
        options = []
        for col in df.columns:
            options.append({"label": col, "value": col})

        return options

    raise PreventUpdate



import base64
import io
import pandas as pd

from dash import html


def load_csv(contents, filename):
    content_type, content_string = contents.split(',')
    decoded = base64.b64decode(content_string)
    try:
        if 'csv' in filename:
            df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))
        elif 'xls' in filename:
            df = pd.read_excel(io.BytesIO(decoded))
        else:
            return html.Div(['Invalid file type. Only csv.']), None
    except Exception as e:
        print(e)
        return html.Div(['There was an error processing this file.']), None

    df = reduce_data(df)
    return html.Div(["Data successfully loaded."]), df


def reduce_data(df: pd.DataFrame) -> pd.DataFrame:
    return df.iloc[::10, :]



from typing import Union

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from chem_analysis.analysis.base_obj.signal_ import Signal
from chem_analysis.analysis.utils.plot_format import get_plot_color, add_plot_format, get_multi_y_axis


class Chromatogram:
    """
    A grouping of Signals.
    """

    __count = 0
    _signal = Signal

    def __init__(self, data: Union[pd.DataFrame, Signal, list[Signal]], name: str = None):
        super().__init__()
        if isinstance(data, pd.DataFrame):
            data = self._load_from_df(data)
        elif isinstance(data, Signal):
            data = {data.y_label if data.y_label is not None else "y_axis": data}
        elif isinstance(data, list):
            if all(isinstance(dat, Signal) for dat in data):
                data = {dat.y_label if dat.y_label is not None else f"y_axis{i}": dat for i, dat in enumerate(data)}
            else:
                raise ValueError("Invalid type in list")

        self.signals = data
        for sig in self.signals:
            setattr(self, sig.name, sig)

        if name is None:
            name = f"Chromat_{Chromatogram.__count}"
            Chromatogram.__count += 1
        self.name = name

    def __repr__(self) -> str:
        text = f"{self.name}: "
        text += "; ".join(self.names)
        return text

    def __iter__(self) -> Signal:
        for sig in self.signals:
            yield sig

    @property
    def names(self):
        return [i.name for i in self.signals]

    @property
    def y_labels(self):
        return [i.y_label for i in self.signals]

    @property
    def x_label(self):
        return self.signals[0].x_label

    @property
    def num_signals(self):
        return len(self.signals)

    def _load_from_df(self, df: pd.DataFrame) -> list[Signal]:
        """ Converts pandas dataframe into traces. """
        signals = []
        for col in df.columns:
            signals.append(self._signal(ser=df[col], _parent=self))

        return signals

    def to_dataframe(self):
        pass

    def baseline(self, **kwargs):
        for sig in self:
            sig.baseline(**kwargs)

    def despike(self, **kwargs):
        for sig in self:
            sig.despike(**kwargs)

    def smooth(self, **kwargs):
        for sig in self:
            sig.smooth(**kwargs)

    def auto_peak_picking(self, **kwargs):
        for sig in self:
            sig.auto_peak_picking(**kwargs)

    def auto_peak_baseline(self, **kwargs):
        for sig in self:
            sig.auto_peak_baseline(**kwargs)

    def stats(self, op_print: bool = True) -> str:
        text = ""
        for i, sig in enumerate(self):
            if i == 0:
                text += sig.stats(op_print=False)
                continue

            text += sig.stats(op_print=False, op_headers=False)

        if op_print:
            print(text)
        return text

    def plot(self, fig: go.Figure = None, auto_open: bool = True, auto_format: bool = True,
             op_peaks: bool = True, **kwargs) -> go.Figure:
        if fig is None:
            fig = go.Figure()

        colors = get_plot_color(self.num_signals)

        for sig, color in zip(self, colors):
            kwargs_ = {"color": color}
            if kwargs:
                kwargs_ = {**kwargs_, **kwargs}
            fig = sig.plot_add_on(fig, auto_open=False, auto_format=False, op_peaks=op_peaks, **kwargs_)

        if auto_format:
            add_plot_format(fig, self.x_label, "; ".join(self.y_labels))

        if auto_open:
            fig.write_html(f'temp.html', auto_open=True)

        return fig

    def plot_sep_y(self, fig: go.Figure = None, auto_open: bool = True,
                   op_peaks: bool = True, spread: float = None, **kwargs) -> go.Figure:
        """ Basic plotting """
        if fig is None:
            fig = go.Figure()

        # generate lines
        colors = get_plot_color(self.num_signals)
        y_axis_labels = self._get_y_labels()

        for sig, color, label in zip(self, colors, y_axis_labels):
            kwargs_ = {"color": color}
            if kwargs:
                kwargs_ = {**kwargs_, **kwargs}

            fig = sig.plot_add_on(fig, auto_open=False, auto_format=False, op_peaks=op_peaks, y_label=label, **kwargs_)

        # adding multiple y-axis
        if spread is None:
            spread = 0.05 * len(set(y_axis_labels))
        axis_format = get_multi_y_axis(colors, fig, spread)
        fig.update_layout(**axis_format)

        if auto_open:
            fig.write_html("temp.html", auto_open=True)

        return fig

    def _get_y_labels(self) -> list[str]:
        y_labels = []
        seen = []
        count = 1
        for sig in self:
            if sig.name[:2] in seen:
                index = [i[:2] for i in seen].index(sig.name[:2])
                y_labels.append(y_labels[index])
            else:
                y_labels.append(f"y{count}")
                count += 1

        return y_labels


def local_run():
    from scipy.stats import norm
    nx = 1000
    ny = 3
    x = np.linspace(0, nx, nx)
    y = np.empty((ny, nx))
    for i in range(ny):
        rv = norm(loc=nx * np.random.random(1), scale=10)
        y[i, :] = np.linspace(0, nx, nx) + 20 * np.random.random(nx) * np.random.random(1) + \
                  5000 * rv.pdf(x) * np.random.random(1)

    df = pd.DataFrame(data=y.T, index=x)
    df.columns = ["RI", "UV", "LS"]
    df.index.names = ["time"]
    chro = Chromatogram(df)
    chro.baseline(deg=1)
    chro.auto_peak_picking()
    chro.plot()
    chro.stats()
    print("done")


if __name__ == "__main__":
    local_run()



from functools import wraps

from scipy.signal import find_peaks


@wraps(find_peaks)
def scipy_find_peaks(y, *args, **kwargs):
    peaks, _ = find_peaks(y, *args, **kwargs)
    return peaks


peak_picking_methods = {"scipy", scipy_find_peaks}



from flask import Flask
from dash import Dash

server = Flask('chem_analysis')
app = Dash(__name__, server=server)

app.config.suppress_callback_exceptions = True
app.css.config.serve_locally = True
app.scripts.config.serve_locally = True



from typing import Union, Callable
from enum import Enum

import numpy as np
import pandas as pd
import plotly.graph_objs as go


from chem_analysis.analysis.base_obj.peak import Peak
from chem_analysis.analysis.algorithms import despike_methods, baseline_methods, smoothing_methods
import chem_analysis.analysis.algorithms as algorithms
from chem_analysis.analysis.logger import logger_analysis
from chem_analysis.analysis.utils import ObjList, fig_count
from chem_analysis.analysis.utils.plot_format import get_plot_color, get_similar_color, add_plot_format

array_like = (np.ndarray, list, tuple)


class ProcessType(Enum):
    RAW = 0
    DESPIKE = 1
    BASELINE = 2
    SMOOTH = 3


class ProcessStep:
    def __init__(self, func: Callable, type_: ProcessType, kwargs: dict = None):
        self.func = func
        self.current_type = type_
        self.kwargs = kwargs

    def __repr__(self):
        return f"{self.func.__name__}; type:{self.current_type}"

    def __call__(self, *args, **kwargs):
        if kwargs is not None:
            kwargs = {**kwargs, **self.kwargs}

        return self.func(*args, **kwargs)


class Signal:
    """ signal

    A signal is any x-y data.

    Attributes
    ----------
    name: str
        Any name the user wants to add.
    raw: pd.Series
        raw data
    x_label: str
        x-axis label
    y_label: str
        y-axis label
    pipeline:
        data processing pipeline
    peaks: Peak
        peaks found in signal
    num_peaks: int
        number of peaks
    result: pd.Series
        data process through pipeline
    result_norm: pd.Series
        data process through pipeline normalized (peak max = 1)

    """
    __count = 0
    _peak = Peak

    def __init__(self,
                 name: str = None,
                 ser: pd.Series = None,
                 x: Union[np.ndarray, pd.Series] = None,
                 y: np.ndarray = None,
                 x_label: str = None,
                 y_label: str = None,
                 _parent=None
                 ):
        """

        Parameters
        ----------
        name: str
            user defined name
        ser: pd.Series
            x-y data in the form of a pandas Series
        x: np.ndarray
            x data
        y: np.ndarray
            y data
        x_label: str
            x-axis label
        y_label: str
            y-axis label

        Notes
        -----
        * Either 'ser' or 'x' and 'y' are required but not both.

        """
        if name is None:
            name = f"trace_{Signal.__count}"
            Signal.__count += 1

        self.name = name
        self.x_label = x_label
        self.y_label = y_label

        if ser is not None and isinstance(ser, pd.Series) and x is None and y is None:
            self.raw = ser
            if self.x_label is None:
                self.x_label = self.raw.index.name
            if self.y_label is None:
                self.y_label = self.raw.name
        elif ser is None and isinstance(x, array_like) and isinstance(y, array_like):
            self.raw = pd.Series(data=y, index=x, name=self.y_label)
            self.raw.index.names = [self.x_label]

        if not hasattr(self, "raw"):
            mes = "Provide either a pandas Series (ser=) or two numpy arrays x and y (x=,y=)."
            raise ValueError(mes + f" (df:{type(ser)}, x:{type(x)}, y:{type(y)})")

        if self.x_label is None:
            self.x_label = "x_axis"
        if self.y_label is None:
            self.y_label = "y_axis"

        self.pipeline = ObjList(ProcessStep)
        self.peaks = ObjList(self._peak)
        self._result = None
        self._result_norm = None
        self._result_up_to_date = False
        self._parent = _parent

    def __repr__(self):
        text = f"{self.name}: "
        text += f"{self.x_label} vs {self.y_label}"
        text += f" (pts: {len(self.raw)})"
        return text

    @property
    def result(self) -> pd.Series:
        if not self._result_up_to_date:
            self.calc()
        return self._result

    @property
    def result_norm(self) -> pd.Series:
        if not self._result_up_to_date:
            self.calc()
        return self._result_norm

    @property
    def num_peaks(self) -> int:
        return len(self.peaks)

    def calc(self):
        x = self.raw.index.to_numpy()
        y = self.raw.to_numpy()
        for func in self.pipeline:
            x, y, z = func(x, y)

        self._result = pd.Series(y, x, name=self.y_label)
        self._result_norm = self._result/np.max(self._result)
        self._result.index.names = [self.x_label]
        self._result_up_to_date = True

    def despike(self, method="default", **kwargs):
        if callable(method):
            func = method
        else:
            try:
                func: callable = despike_methods[method]
            except KeyError:
                raise ValueError(f"Not a valid 'despiking' method. (given: {method})")

        self.pipeline.add(ProcessStep(func, ProcessType.DESPIKE, kwargs))
        self._result_up_to_date = False

        logger_analysis.debug(f"Despiking ({method}) done on: '{self.name}'.")

    def baseline(self, method="polynomial", **kwargs):
        if callable(method):
            func = method
        else:
            try:
                func: callable = baseline_methods[method]
            except KeyError:
                raise ValueError(f"Not a valid 'baseline' method. (given: {method})")

        self.pipeline.add(ProcessStep(func, ProcessType.BASELINE, kwargs))
        self._result_up_to_date = False

        logger_analysis.debug(f"Baseline correction ({method}) done on: '{self.name}'.")

    def smooth(self, method="default", **kwargs):
        if callable(method):
            func = method
        else:
            try:
                func: callable = smoothing_methods[method]
            except KeyError:
                raise ValueError(f"Not a valid 'smooth' method. (given: {method})")

        self.pipeline.add(ProcessStep(func, ProcessType.SMOOTH, kwargs))
        self._result_up_to_date = False

        logger_analysis.debug(f"Smooth ({method}) done on: '{self.name}'.")

    def auto_peak_picking(self, limit_range: list[float] = None, **kwargs):
        self.peaks.clear()

        # find peaks
        kwargs_ = {"width": self.raw.index[-1] / 100, "height": 0.03, "prominence": 0.03}
        if kwargs:
            kwargs_ = {**kwargs_, **kwargs}
        if limit_range:
            lb_index = np.argmin(np.abs(self.result.index.to_numpy() - limit_range[0]))
            ub_index = np.argmin(np.abs(self.result.index.to_numpy() - limit_range[1]))
            y = self.result_norm.iloc[lb_index:ub_index].to_numpy()
            y = y/np.max(y)
            peaks_index = algorithms.scipy_find_peaks(y, **kwargs_) + lb_index
        else:
            peaks_index = algorithms.scipy_find_peaks(self.result_norm.to_numpy(), **kwargs_)

        # get bounds from peak maximums
        if len(peaks_index) != 0:
            for peak in peaks_index:
                lb, ub = algorithms.rolling_value(self.result_norm.to_numpy(), peak_index=peak, sensitivity=0.1,
                                                  cut_off=0.05)
                self.peaks.add(self._peak(self, lb, ub))
        else:
            logger_analysis.warning(f"No peaks found in signal '{self.name}'.")

        logger_analysis.debug(f"Auto peak picking done on: '{self.name}'. Peaks found: {self.num_peaks}")

    def auto_peak_baseline(self, iterations: int = 3, limit_range: list[float] = None, **kwargs):
        """
        Does automatic baseline correction and peak detection.
        It uses peak detection to create a mask for baseline correction in an iterative fashion.

        """
        self.baseline(**kwargs)
        self.auto_peak_picking(limit_range=limit_range)
        for i in range(iterations):
            self.pipeline.remove(-1)
            mask = np.ones_like(self.raw.to_numpy())
            for peak in self.peaks:
                mask[peak.slice] = False
            self.baseline(mask=mask, **kwargs)
            self.auto_peak_picking()
            logger_analysis.debug(f"'auto_peak_baseline' iteration {i} done.")

    def stats(self, op_print: bool = True, op_headers: bool = True) -> str:
        """ Print out signal/peak stats. """
        text = ""
        for i, peak in enumerate(self.peaks):
            if i == 0:
                if op_headers:
                    text += peak.stats(op_print=False)
                    continue

            text += peak.stats(op_print=False, op_headers=False)

        if op_print:
            print(text)
        return text

    def plot(self, fig: go.Figure = None, auto_open: bool = True, auto_format: bool = True,
             op_peaks: bool = True, y_label: str = None, title: str = None, **kwargs) -> go.Figure:
        """ Plot

        General plotting

        Parameters
        ----------
        fig: go.Figure
            plotly figure; will automatically create if not provided
        auto_open: bool
            create "temp.html" and auto_open in browser
        auto_format: bool
            apply built-in formatting
        op_peaks: bool
            add peak plotting stuff
        y_label: str
            y_axis label (used for multiple y-axis)
        title: str
            title

        Returns
        -------
        fig: go.Figure
            plotly figure

        """
        if fig is None:
            fig = go.Figure()

        if "color" in kwargs:
            color = kwargs.pop("color")
        else:
            color = 'rgb(0,0,0)'

        group = self.name

        # add peaks
        if op_peaks:
            if len(self.peaks) > 0:
                if color == 'rgb(0,0,0)':
                    peak_color = get_plot_color(self.num_peaks)
                else:
                    peak_color = get_similar_color(color, self.num_peaks)

                for peak, color_ in zip(self.peaks, peak_color):
                    peak.plot_add_on(fig, color=color_, group=group, y_label=y_label)

        # add main trace
        plot_kwargs = {
            "x": self.result.index,
            "y": self.result,
            "mode": 'lines',
            "connectgaps": True,
            "name": self.result.name,
            "legendgroup": group,
            "line": dict(color=color)
        }
        if y_label is not None:
            plot_kwargs["yaxis"] = y_label

        fig.add_trace(go.Scatter(**plot_kwargs))

        if auto_format:
            if title is not None:
                fig.update_layout(title=title)
            add_plot_format(fig, self.result.index.name, str(self.result.name))

        if auto_open:
            global fig_count
            fig.write_html(f'temp{fig_count}.html', auto_open=True)
            fig_count += 1

        return fig


def local_run():
    from scipy.stats import norm
    n = 1000
    rv = norm(loc=n/2, scale=10)
    x = np.linspace(0, n, n)
    y = np.linspace(0, n, n) + 20 * np.random.random(n) + 5000 * rv.pdf(x)
    signal = Signal(name="test", x=x, y=y, x_label="x_test", y_label="y_test")
    signal.baseline(deg=1)
    signal.auto_peak_picking()
    signal.stats()
    signal.plot()
    print("done")


if __name__ == '__main__':
    local_run()



def parse_label(text: str) -> Union[tuple[str, str], tuple[str, None]]:
    """ find units in parentheses """
    list_text = re.split("(.*)\((.*)\)(.*)", text)
    list_text = [text for text in list_text if text]
    if len(list_text) == 1:
        return list_text[0], None
    if len(list_text) == 2:
        return list_text[0], list_text[1]
    else:
        return text, None



import numpy as np
from numpy import ndarray


def detect_outliers(data: np.ndarray, m: float = 2):
    dist_from_median = np.abs(data - np.median(data))
    median_deviation = np.median(dist_from_median)
    if median_deviation != 0:
        scale_distances_from_median = dist_from_median/median_deviation
        return scale_distances_from_median > m  # True is an outlier

    return np.zeros_like(data)  # no outliers


def window_calc(data: np.ndarray, pos: int, m: float = 2) -> float:
    outlier = detect_outliers(data, m)
    if outlier[pos]:
        return np.median(data[np.invert(outlier)])
    else:
        return data[pos]


def despike(x: np.ndarray, y: np.ndarray, window: int = 20, m: float = 2) -> tuple[ndarray, ndarray]:
    out = np.empty_like(y)

    if window % 2 != 0:
        window += 1

    span = int(window/2)
    for i in range(len(y)):
        if i < span:  # left edge
            out[i] = window_calc(y[:window], i, m)
        elif i > len(y) - span:  # right edge
            out[i] = window_calc(y[window:], i - (len(y)-window), m)
        else:  # middle
            out[i] = window_calc(y[i-span:i+span], span, m)

    return x, out


despike_methods = {"default": despike}


def test():
    import plotly.graph_objs as go

    n = 1000
    x = np.linspace(0, n-1, n)
    y = np.ones(n) + np.random.random(n)
    y[100] = 5
    y[101] = 3

    y[500] = 2
    y[501] = 1.7
    y[502] = 1.5
    y[503] = 1.2

    y[700] = 0.5
    y[701] = 0
    y[702] = -0.5
    y[703] = -1

    x, y_new = despike(x, y)

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=x, y=y, mode="lines"))
    fig.add_trace(go.Scatter(x=x, y=y_new, mode="lines"))
    fig.write_html("temp.html", auto_open=True)


if __name__ == "__main__":
    test()



from chem_analysis.server import app

from dash import html, dcc
from dash.dependencies import Input, Output

# layouts
app.layout = html.Div([
    html.H1('Size Exclusion Analysis'),
    dcc.Tabs(
        children=[
            dcc.Tab(label='Load', value='tab-load'),
            dcc.Tab(label='Process', value='tab-process'),
            dcc.Tab(label='Analysis', value='tab-analysis')
        ],
        value='tab-Load',
        id='tabs'
    ),
    html.Div(id='tabs-content'),

    dcc.Store(id='dataframe'),
    dcc.Store(id='tab_load_reload'),
])

from chem_analysis.tabs.tab_load import tab_load_layout
from chem_analysis.tabs.tab_process import tab_process_layout
from chem_analysis.tabs.tab_analysis import tab_analysis_layout


# callbacks
@app.callback(Output('tabs-content', 'children'),
              [Input('tabs', 'value')])
def render_content(tab):
    if tab == 'tab-load':
        return tab_load_layout
    elif tab == 'tab-process':
        return tab_process_layout
    elif tab == 'tab-analysis':
        return tab_analysis_layout

    return tab_load_layout

from chem_analysis.tabs.tab_load_callback import *
from chem_analysis.tabs.tab_process_callback import *
from chem_analysis.tabs.tab_analysis_callback import *



from enum import Enum
from typing import Union, Callable

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from chem_analysis.analysis.base_obj.calibration import Cal
from chem_analysis.analysis.base_obj.signal_ import Signal
from chem_analysis.analysis.sec.sec_peak import SECPeak
from chem_analysis.analysis.utils.plot_format import add_plot_format
from chem_analysis.analysis.utils import fig_count


class SECType(Enum):
    RI = 0
    UV = 1
    LS = 2
    VISC = 3


class SECSignal(Signal):
    """ SECSignal

    Extends Signal for SEC (size exclusion chronograph)

    Attributes
    ----------
    cal: Cal
        calibration
    peaks: List[SECPeak]


    """

    _peak = SECPeak

    def __init__(self, cal: Union[Cal, Callable] = None, **kwrags):
        super().__init__(**kwrags)

        if cal is None and self._parent.cal is not None:
            cal = self._parent.cal
        if not isinstance(cal, Cal):
            cal = Cal(cal)
        self.cal = cal

        self._result_weight = None

    @property
    def result_weight(self) -> pd.Series:
        if not self._result_up_to_date:
            self.calc()
        return self._result_weight

    def calc(self):
        super().calc()

        mol_weight = self.cal(self.result.index.to_numpy())
        self._result_weight = pd.Series(self.result.to_numpy(), mol_weight)

    def auto_peak_baseline(self, iterations: int = 3, limit_range: list[float] = None, **kwargs):
        # if limit_range is None and self.cal.lb and self.cal.ub:
        #     limit_range = [self.cal.ub_loc, self.cal.lb_loc]

        super().auto_peak_baseline(iterations=iterations, limit_range=limit_range, **kwargs)

    def plot(self, fig: go.Figure = None, auto_open: bool = True, auto_format: bool = True,
             op_peaks: bool = True, op_cal: bool = True, y_label: str = None, title: str = None, **kwargs) -> go.Figure:
        if fig is None:
            fig = go.Figure()

        fig = super().plot(fig, auto_open=False, op_peaks=op_peaks, y_label=y_label)

        if op_cal and self.cal is not None:
            self._plot_cal(fig)

        if auto_format:
            if title is not None:
                fig.update_layout(title=title)
            add_plot_format(fig, self.result.index.name, str(self.result.name))

        if auto_open:
            global fig_count
            fig.write_html(f'temp{fig_count}.html', auto_open=True)
            fig_count += 1

        return fig

    def _plot_cal(self, fig: go.Figure, **kwargs):
        color = "rgb(130,130,130)"
        kwargs_ = {
            "width": 1,
            "color": color
        }
        if kwargs:
            kwargs_ = {**kwargs_, **kwargs}

        kkwargs = {}

        mw = self.cal(self.result.index.to_numpy())
        time = self.result.index
        fig.add_trace(go.Scatter(
            x=time,
            y=mw,
            name=self.cal.name if self.cal.name is not None else "calibration",
            mode="lines",
            line=kwargs_,
            yaxis="y2",
            showlegend=True,
            legendgroup="cal",
            **kkwargs
        ))

        kwargs_["dash"] = 'dash'

        # low limit
        if self.cal.lb:
            fig.add_trace(go.Scatter(
                x=[self.cal.lb_loc, self.cal.lb_loc],
                y=[0, np.max(mw)],
                name="calibration",
                mode="lines",
                line=kwargs_,
                yaxis="y2",
                showlegend=False,
                legendgroup="cal",
                **kkwargs
            ))

        # up limit
        if self.cal.ub:
            fig.add_trace(go.Scatter(
                x=[self.cal.ub_loc, self.cal.ub_loc],
                y=[0, np.max(mw)],
                name="calibration",
                mode="lines",
                line=kwargs_,
                yaxis="y2",
                showlegend=False,
                legendgroup="cal",
                **kkwargs
            ))

        fig.update_layout(
            yaxis2=dict(
                title="molecular weight",
                titlefont=dict(
                    color=color
                ),
                tickfont=dict(
                    color=color
                ),
                anchor="x",
                overlaying="y",
                side="right",
                type="log",
                range=[2, 6]
            ),
        )


def local_run():
    from scipy.stats import norm

    def cal_func(time: np.ndarray):
        return 10**(0.0167 * time ** 2 - 0.9225 * time + 14.087)

    cal = Cal(cal_func, lb=900, ub=319_000)

    nx = 1000
    x = np.linspace(0, 25, nx)
    rv = norm(loc=15, scale=0.6)
    rv2 = norm(loc=18, scale=0.6)
    y = 5 * np.linspace(0, 1, nx) + np.random.random(nx) + 100 * rv.pdf(x) + 20 * rv2.pdf(x)

    sig = SECSignal(name="RI data", x=x, y=y, x_label="time (min)", y_label="intensity", cal=cal)
    sig.auto_peak_baseline(deg=1)
    sig.plot()
    sig.stats()
    print("done")


if __name__ == '__main__':
    local_run()



from chem_analysis.analysis.algorithms.despike import despike_methods
from chem_analysis.analysis.algorithms.baseline import baseline_methods
from chem_analysis.analysis.algorithms.smoothing import smoothing_methods
from chem_analysis.analysis.algorithms.peak_picking import peak_picking_methods, scipy_find_peaks
from chem_analysis.analysis.algorithms.bound_detection import bound_detection_methods, rolling_value



import math

import numpy as np
import plotly.graph_objs as go

from chem_analysis.analysis.utils.plot_format import add_plot_format


class PolyLogNorm:
    _count = 0

    def __init__(self, Mn: float = 10_000, D: float = 1.1, Mw: float = None, m_i: np.ndarray = None,
                 name: str = None):
        if 0 >= Mn:
            raise ValueError("Mn must be positive.")
        if Mw is not None and 0 >= Mw:
            raise ValueError("Mw must be positive.")
        if 1 >= D:
            raise ValueError("D must be greater than  1.")

        self.Mn = Mn
        if Mw:
            self.Mw = Mw
            self.D = Mw / Mn
        else:
            self.D = D
            self.Mw = Mn * D

        self.ln_sigma = math.sqrt(math.log(D))
        self.ln_mean = math.log(self.Mn) - self.ln_sigma ** 2 / 2

        if m_i is None:
            m_i = np.logspace(2, 6, 1000)
        self.m_i = m_i

        if name is None:
            self.name = f"log_normal_{self._count}"
            PolyLogNorm._count += 1

    def __repr__(self) -> str:
        return f"Mn: {self.Mn}; D: {self.D}"

    def __call__(self, m_i: np.ndarray = None) -> np.ndarray:
        return self.weight(m_i)

    @property
    def mole(self, m_i: np.ndarray = None) -> np.ndarray:
        if m_i is None:
            m_i = self.m_i
        return 1 / (m_i * np.sqrt(2 * np.pi * np.log(self.D))) * \
               np.exp(-(np.log(m_i / self.Mn) + np.log(self.D) / 2) ** 2 / (2 * np.log(self.D)))

    @property
    def weight(self, m_i: np.ndarray = None) -> np.ndarray:
        if m_i is None:
            m_i = self.m_i
        return 1 / (self.Mn * np.sqrt(2 * np.pi * np.log(self.D))) * \
               np.exp(-(np.log(m_i / self.Mn) + np.log(self.D) / 2) ** 2 / (2 * np.log(self.D)))

    def plot(self, fig: go.Figure = None, op_type: str = "weight", op_log: bool = True,
             auto_open: bool = True, auto_format: bool = True, **kwargs) -> go.Figure:
        """ Basic plotting """
        if fig is None:
            fig = go.Figure()

        if "color" in kwargs:
            color = kwargs.pop("color")
        else:
            color = 'rgb(0,0,0)'

        # add main trace
        plot_kwargs = {
            "mode": 'lines',
            "connectgaps": True,
            "name": self.name,
            "line": dict(color=color)
        }
        if op_type == "weight":
            plot_kwargs["x"] = self.m_i
            plot_kwargs["y"] = self.weight
            y_axis = "weight fraction"
        elif op_type == "mole":
            plot_kwargs["x"] = self.m_i
            plot_kwargs["y"] = self.mole
            y_axis = "mole fraction"
        else:
            raise ValueError(f"Invalid 'op_type'. (given: {op_type})")

        fig.add_trace(go.Scatter(**plot_kwargs))

        if auto_format:
            if op_log:
                format_kwargs = {"type": "log"}
            else:
                format_kwargs = {}
            add_plot_format(fig, "molecular weight", y_axis, x_kwargs=format_kwargs)

        if auto_open:
            fig.write_html(f'temp.html', auto_open=True)

        return fig


def run_local():
    dis = PolyLogNorm()
    dis.plot()


if __name__ == '__main__':
    run_local()



import os
import glob
import datetime


def check_if_recent(folder: str, pattern: str = "temp*") -> int:
    os.chdir(folder)
    files = glob.glob(pattern)
    now = datetime.datetime.now()
    times = [os.path.getmtime(file) - now for file in files]

    for file, time in zip(files, times):
        if time < 1:
            return file
    else:
        return file


from typing import Protocol

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from chem_analysis.analysis.utils.plot_format import add_plot_format, get_multi_y_axis, get_plot_color
from chem_analysis.analysis.base_obj.calibration import Cal
from chem_analysis.analysis.base_obj.peak import Peak
from chem_analysis.analysis.utils import fig_count


def cal_Mn_D_from_wi(mw_i: np.ndarray, wi: np.ndarray) -> tuple[float, float]:
    """ calculate Mn and D from wi vs MW data (MW goes low to high) """
    data_points = len(mw_i)

    # flip data if giving backwards
    if mw_i[1] > mw_i[-1]:
        mw_i = np.flip(mw_i)
        wi = np.flip(wi)

    wi_d_mi = np.zeros(data_points)
    wi_m_mi = np.zeros(data_points)
    for i in range(data_points):
        if mw_i[i] != 0:
            wi_d_mi[i] = wi[i] / mw_i[i]
        wi_m_mi[i] = wi[i] * mw_i[i]

    mw_n = np.sum(wi) / np.sum(wi_d_mi)
    mw_w = np.sum(wi_m_mi) / np.sum(wi)
    mw_d = mw_w / mw_n
    return mw_n, mw_d


class SECPeakSupports(Protocol):
    name: str
    result: pd.Series
    result_weight: pd.Series
    cal: Cal


class SECPeak(Peak):
    """ SECPeak

    Extends the Peak class to for SEC (size exclusion chromatograph).

    Attributes
    ----------
    mw_i: np.ndarray
        molecular weight of i-mer
    wi: np.ndarray
        weight fraction of i-mer
    xi: np.ndarray
        mole fraction of i-mer
    mw_n: float
        number average molecular weight
    mw_w: float
        weight average molecular weight
    mw_max: float
        molecular weight at peak max
    mw_d: float
        dispersity of molecular weight
    mw_mean: float
        mean molecular weight distribution (same as mw_n)
    mw_std: float
        standard deviation of molecular weight distribution
    mw_skew: float
        skew of molecular weight distribution
        symmetric: -0.5 to 0.5; moderate skew: -1 to -0.5 or 0.5 to 1; high skew: <-1 or >1;
        positive tailing to higher numbers; negative tailing to smaller numbers
    mw_kurtosis: float
        kurtosis of molecular weight distribution (Fisher)  (Warning: highly sensitive to peak bounds)
        negative: flatter peak; positive: sharp peak
    mw_fwhm: float
        full width half max of molecular weight distribution
    mw_asym: float
        asymmetry factor of molecular weight distribution; distance from the center line of the peak to the back
        slope divided by the distance from the center line of the peak to the front slope
        >1 tailing to larger values; <1 tailing to smaller numbers

    """

    def __init__(self, parent: SECPeakSupports, lb_index: int, hb_index: int):
        """

        Parameters
        ----------
        parent: SECPeakSupports
            parent object that the peak is associated with
        lb_index: int
            lower bound index
        hb_index: int
            higher bound index

        """
        self.mw_i = None
        self.wi = None
        self.xi = None

        self.mw_n = None
        self.mw_w = None
        self.mw_max = None
        self.mw_d = None
        self.mw_mean = None
        self.mw_std = None
        self.mw_skew = None
        self.mw_kurtosis = None
        self.mw_fwhm = None
        self.mw_asym = None

        super().__init__(parent, lb_index, hb_index)

    def __repr__(self):
        return f"peak: {self.id_} at {self.max_loc} with Mn: {self.mw_n} (D: {self.mw_d})"

    def calc(self):
        """ Calculates the molecular weight stats for the peak. """
        super().calc()

        self.mw_i = np.flip(self._parent.result_weight.index[self.slice].to_numpy())
        # 'np.flip' so small mw is first avoids having to flip everything in calc below
        self.wi = np.flip(self._parent.result_weight.iloc[self.slice].to_numpy() / \
                  np.trapz(x=self.mw_i, y=self._parent.result_weight.iloc[self.slice].to_numpy()))

        self.mw_n, self.mw_d = cal_Mn_D_from_wi(mw_i=self.mw_i, wi=self.wi)
        self.mw_w = self.mw_n * self.mw_d
        self.mw_max = self._parent.cal(self.max_loc)

        self.xi = self.wi * self.mw_n / self.mw_i

        self.mw_mean = np.trapz(x=self.mw_i, y=self.mw_i*self.xi)
        self.mw_std = np.sqrt(np.trapz(x=self.mw_i, y=self.xi * (self.mw_i - self.mw_mean) ** 2))
        self.mw_skew = np.trapz(x=self.mw_i, y=self.xi * (self.mw_i - self.mw_mean) ** 3) / self.mw_std ** 3
        self.mw_kurtosis = (np.trapz(x=self.mw_i, y=self.xi * (self.mw_i - self.mw_mean) ** 4) / self.mw_std ** 4)- 3
        self.mw_fwhm = self.get_fw(x=self.mw_i, y=self.xi, height=0.5)
        self.mw_asym = self.get_asym(x=self.mw_i, y=self.xi, height=0.1)

    def plot_mw(self, fig: go.Figure = None, auto_open: bool = True, auto_format: bool = True, y_label: str = None,
                title: str = None, spread: float = None, **kwargs) -> go.Figure:
        if fig is None:
            fig = go.Figure()

        colors = get_plot_color(2)

        # add main trace
        plot_kwargs_xi = {
            "x": self.mw_i,
            "y": self.xi,
            "mode": 'lines',
            "connectgaps": True,
            "name": "xi",
            "line": dict(color=colors[0]),
            "yaxis": "y1"
        }

        fig.add_trace(go.Scatter(**plot_kwargs_xi))
        plot_kwargs_wi = {
            "x": self.mw_i,
            "y": self.wi,
            "mode": 'lines',
            "connectgaps": True,
            "name": "wi",
            "line": dict(color=colors[1]),
            "yaxis": "y2"
        }

        fig.add_trace(go.Scatter(**plot_kwargs_wi))

        # adding multiple y-axis
        y_axis_labels = ["y1", "y2"]
        if spread is None:
            spread = 0.05 * len(set(y_axis_labels))
        axis_format = get_multi_y_axis(colors, fig, spread)
        fig.update_layout(**axis_format)

        if auto_open:
            global fig_count
            fig.write_html(f'temp{fig_count}.html', auto_open=True)
            fig_count += 1

        return fig

    def stats(self, op_print: bool = True, op_headers: bool = True, window: int = 150, headers: dict = None):
        if headers is None:
            headers = {  # attribute: print
                "id_": "id", "lb_loc": "low bound", "max_loc": "max", "hb_loc": "high bound", "area": "area",
                "mw_n": "mw_n", "mw_w": "mw_w", "mw_d": "mw_d", "mw_max": "mw_max", "mw_std": "mw_std",
                "mw_skew": "mw_skew", "mw_kurtosis": "mw_kurtosis", "mw_asym": "mw_asym"
            }

        return super().stats(op_print, op_headers, window, headers)



"""
Logger

"""

import logging
from functools import wraps


color_codes = {
    "reset": "\033[0m",  # add at the end to stop coloring
    "black": "\033[30m",
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "white": "\033[37m",
}

BOLD_SEQ = "\033[1m"


class LogFormatter(logging.Formatter):
    def __init__(self):
        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=None, style='%')

    def format(self, record):
        # Save the original format configured by the user when the logger formatter was instantiated
        format_orig = self._style._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.DEBUG:
            self._style._fmt = f"\t \t{color_codes['white']}DEBUG: %(msg)s{color_codes['reset']}"
        elif record.levelno == logging.INFO:
            self._style._fmt = f"\t{color_codes['white']}INFO: %(msg)s{color_codes['reset']}"
        elif record.levelno == logging.WARNING:
            self._style._fmt = f"{color_codes['yellow']}WARNING: %(msg)s {color_codes['reset']}"
        elif record.levelno == logging.ERROR:
            self._style._fmt = f"{color_codes['red']}ERROR: %(msg)s {color_codes['reset']}"
        elif record.levelno == logging.CRITICAL:
            self._style._fmt = f"{color_codes['red']}CRITICAL: %(msg)s {color_codes['reset']}"

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = format_orig

        return result


logger_analysis = logging.getLogger("analysis")

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(LogFormatter())
logger_analysis.addHandler(stream_handler)

logger_analysis.setLevel(logging.DEBUG)


# Logger decorators
# def log_debug(func):
#     """ Add 'input --> output' logging to a function. At DEBUG level. """
#     @wraps(func)
#     def _log_debug(*args, **kwargs):
#         result = func(*args, **kwargs)
#         logger.debug(f"{func.__name__}: {args} --> {result}")
#         return result
#     return _log_debug
#
#
# def log_info(func):
#     """ Add 'input --> output' logging to a function. At INFO level. """
#     @wraps(func)
#     def _log_info(*args, **kwargs):
#         result = func(*args, **kwargs)
#         logger.info(f"{func.__name__}: {args} --> {result}")
#         return result
#     return _log_info


from typing import Union
import re

import numpy as np
import pandas as pd
import plotly.graph_objs as go


layout_figure = {
        "autosize": False,
        "width": 1200,
        "height": 600,
        "font": dict(family="Arial", size=18, color="black"),
        "plot_bgcolor": "white",
        "legend": {"x": 0.05, "y": 0.95}
    }

layout_xaxis = {
    # "title": "<b>X<b>",
    "tickprefix": "<b>",
    "ticksuffix": "</b>",
    "showline": True,
    "linewidth": 5,
    "mirror": True,
    "linecolor": 'black',
    "ticks": "outside",
    "tickwidth": 4,
    "showgrid": False,
    "gridwidth": 1,
    "gridcolor": 'lightgray'
}

layout_yaxis = {
    # "title": "<b>Y<b>",
    "tickprefix": "<b>",
    "ticksuffix": "</b>",
    "showline": True,
    "linewidth": 5,
    "mirror": True,
    "linecolor": 'black',
    "ticks": "outside",
    "tickwidth": 4,
    "showgrid": False,
    "gridwidth": 1,
    "gridcolor": 'lightgray'
}

hex_options = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']


def get_plot_color(num_colors: int = 1) -> list[str]:
    """Gets color for 2D plots."""
    color_list = [
        'rgb(10,36,204)',  # blue
        'rgb(172,24,25)',  # red
        'rgb(6,127,16)',  # green
        'rgb(251,118,35)',  # orange
        'rgb(145,0,184)',  # purple
        'rgb(255,192,0)'  # yellow
    ]
    if num_colors <= 1:
        return [color_list[0]]
    if num_colors <= len(color_list):
        return color_list[:num_colors]
    else:
        num_colors_extra = num_colors - len(color_list)
        for i in range(num_colors_extra):
            color = ["#" + ''.join([np.random.choice(hex_options) for _ in range(6)])]
            color = [col.lstrip('#') for col in color]
            color = ["rgb" + str(tuple(int(col[i:i + 2], 16) for i in (0, 2, 4))) for col in color]
            color_list = color_list + color

        return color_list


def get_similar_color(color_in: str, num_colors: int, mode: str = "dark") -> list[str]:
    rgb = re.findall("[0-9]{1,3}", color_in)
    rgb = [int(i) for i in rgb]
    if mode == "dark":
        change_rgb = [i > 120 for i in rgb]
        jump_amount = [-int((i - 80) / num_colors) for i in rgb]
        jump_amount = [v if i else 0 for i, v in zip(change_rgb, jump_amount)]

    elif mode == "light":
        jump_amount = [int(100 / num_colors) if i < 100 else int((245-i)/num_colors) for i in rgb]

    else:
        raise ValueError(f"Invalid 'mode'; only 'light' or 'dark'. (mode: {mode})")

    colors = []
    for i in range(num_colors):
        r = rgb[0] + jump_amount[0] * (i+1)
        g = rgb[1] + jump_amount[1] * (i+1)
        b = rgb[2] + jump_amount[2] * (i+1)
        colors.append(f"rgb({r},{g},{b})")

    return colors


def add_plot_format(fig: go.Figure, x_axis: str, y_axis: str, layout_kwargs: dict = None,
                    x_kwargs: dict = None, y_kwargs: dict = None):
    """ Add default format to plot_add_on. """
    if layout_kwargs:
        layout_format = {**layout_figure, **layout_kwargs}
    else:
        layout_format = layout_figure
    fig.update_layout(layout_format)

    # x-axis
    x_axis_format = {"title": x_axis}
    if x_kwargs:
        x_axis_format = {**x_axis_format, **layout_xaxis, **x_kwargs}
    else:
        x_axis_format = {**x_axis_format, **layout_xaxis}
    fig.update_xaxes(x_axis_format)

    # y-axis
    y_axis_format = {"title": y_axis}
    if y_kwargs:
        y_axis_format = {**y_axis_format, **layout_yaxis, **y_kwargs}
    else:
        y_axis_format = {**y_axis_format, **layout_yaxis}
    fig.update_yaxes(y_axis_format)


def get_multi_y_axis(colors: list[str], fig: go.Figure, spread: float = 0.2) -> dict:
    y_axis_labels = {data.yaxis for data in fig.data}
    y_axis_labels = [label for label in y_axis_labels if label is not None]
    num_y_axis = len(y_axis_labels)

    gap = spread / (num_y_axis - 1)
    # first trace
    axis_format = {
        "xaxis": {"domain": [spread, 1]},  # 'tickformat': '%I:%M %p \n %b %d'
        # https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format
        "yaxis1": {
            "title": {"text": f"{y_axis_labels[0]}", "standoff": 0},
            # "tickformat": ".4f",
            # "titlefont": {"color": colors[0]},
            # "tickfont": {"color": colors[0]},
            "tickangle": -45
        }
    }
    # the rest of the traces
    for i in range(1, num_y_axis):
        axis_format[f"yaxis{i + 1}"] = {
            "title": {"text": f"{y_axis_labels[i]}", "standoff": 0},
            # "tickformat": ".4f",
            # "titlefont": {"color": colors[i]},
            # "tickfont": {"color": colors[i]},
            "anchor": "free",
            "overlaying": "y",
            "side": "left",
            "position": spread - gap * i,
            "tickangle": -45
        }

    return axis_format


#######################################################################################################################
#######################################################################################################################
def plot(df: pd.DataFrame, fig=None, auto_open: bool = True, auto_format: bool = True):
    """ Basic plotting """
    if fig is None:
        fig = go.Figure()

    # generate lines
    colors = get_plot_color(len(df.columns))
    traces = get_traces(df, colors)
    for trace in traces:
        fig.add_trace(trace)

    if auto_format:
        add_plot_format(fig, df.index.name, df.columns[0])

    if auto_open:
        fig.write_html("temp.html", auto_open=True)


def get_traces(df: pd.DataFrame, colors: list[str], y_axis_labels: dict = None):
    traces = []
    for i, col in enumerate(df.columns):
        plot_kwargs = {
            "x": df.index,
            "y": df[col],
            "mode": 'lines',
            "connectgaps": True,
            "name": col,
            "line": dict(color=colors[i])
        }
        if y_axis_labels is not None:
            for label in y_axis_labels:
                if col.startswith(label):
                    plot_kwargs["yaxis"] = y_axis_labels[label]
                    break
            else:
                plot_kwargs["yaxis"] = f"y{len(y_axis_labels) + 1}"
                break

        traces.append(go.Scatter(**plot_kwargs))

    return traces


def plot_sep_y(df: pd.DataFrame, fig=None, separate_y_axis=None,
               auto_open: bool = True, auto_format: bool = True, spread=0.20):

    """ Basic plotting """
    if fig is None:
        fig = go.Figure()

    # generate lines
    colors = get_plot_color(len(df.columns))
    traces = get_traces(df, colors, separate_y_axis)
    for trace in traces:
        fig.add_trace(trace)

    # adding multiple y-axis
    axis_format = get_multi_y_axis(colors, fig)
    fig.update_layout(**axis_format)

    if auto_open:
        fig.write_html("temp.html", auto_open=True)

    return fig


def plot_series(data: pd.Series, fig=None, auto_open: bool = True, auto_format: bool = True):
    """ Basic plotting """
    if fig is None:
        fig = go.Figure()

    colors = get_plot_color(1)
    fig.add_trace(
        go.Scatter(x=data.index, y=data, mode="lines", connectgaps=True, name=data.name, line=dict(color=colors)))

    if auto_format:
        add_plot_format(fig, data.index.name, data.name)

    if auto_open:
        fig.write_html(f'temp.html', auto_open=True)

    return fig



import numpy as np


def poly_baseline(x: np.ndarray, y: np.ndarray, deg: int = 0, mask: np.ndarray = None) \
        -> tuple[np.ndarray, np.ndarray, np.ndarray]:

    if mask is not None:
        x_mask = x[np.where(mask)]
        y_mask = y[np.where(mask)]
    else:
        x_mask = x
        y_mask = y

    params = np.polyfit(x_mask, y_mask, deg)
    func_baseline = np.poly1d(params)
    y_baseline = func_baseline(x)
    y = y - y_baseline
    return x, y, y_baseline


baseline_methods = {"polynomial": poly_baseline}



from typing import Protocol, Union

import numpy as np
import pandas as pd
import plotly.graph_objs as go

# from chem_analysis.analysis.utils.plot_format import add_plot_format
from chem_analysis.analysis.utils.sig_fig import sig_figs
from chem_analysis.analysis.logger import logger_analysis
from chem_analysis.analysis.utils import fig_count


class PeakSupports(Protocol):
    name: str
    result: pd.Series


def _length_limit(label: str, limit: int) -> str:
    """ For printing out stats. """
    if len(label) > limit:
        return label[0:limit - 5] + "..."

    return label


class Peak:
    """ Peak

    a peak of a signal

    Attributes
    ----------
    id_: int
        id of peak
    x: nd.ndarray
        x axis data
    y: nd.ndarray
        y axis data
    slice: slice
        slice of full signal for peak
    lb_index: int
        index of lower bound
    hb_index: int
        index of higher bound
    lb_loc: float
        x location of lower bound
    hb_loc: float
        x location of higher bound
    lb_value: float
        y value of lower bound
    hb_value: float
        y value of higher bound
    max_index: int
        index of peak maximum
    max: float
        y value of peak maximum
    max_loc: float
        x location of peak maximum
    area: float
        area under the peak
    mean: float
        average value
    std: float
        standard deviation
    skew: float
        skew
        symmetric: -0.5 to 0.5; moderate skew: -1 to -0.5 or 0.5 to 1; high skew: <-1 or >1;
        positive tailing to higher numbers; negative tailing to smaller numbers
    kurtosis: float
        kurtosis (Fisher) (Warning: highly sensitive to peak bounds)
        negative: flatter peak; positive: sharp peak
    fwhm: float
        full width at half maximum
    asym: float
        asymmetry factor; distance from the center line of the peak to the back slope divided by the distance from the
        center line of the peak to the front slope;
        >1 tailing to larger values; <1 tailing to smaller numbers

    """
    def __init__(self, parent: PeakSupports, lb_index: int, hb_index: int, id_: int = None):
        self.id_ = id_
        self._parent = parent

        self.x = None
        self.y = None
        self.y_norm = None

        self._lb_index = None
        self._hb_index = None
        self.lb_loc = None
        self.hb_loc = None
        self.lb_value = None
        self.hb_value = None
        self.max_index = None
        self.max = None
        self.max_loc = None
        self.area = None

        self.mean = None
        self.std = None
        self.skew = None
        self.kurtosis = None
        self.fwhm = None
        self.asym = None

        self.lb_index = lb_index
        self.hb_index = hb_index

    def __repr__(self):
        return f"peak: {self.id_} at {self.max_loc}"

    @property
    def lb_index(self):
        return self._lb_index

    @lb_index.setter
    def lb_index(self, lb_index):
        self._lb_index = lb_index
        if self._hb_index is not None:
            self.calc()

    @property
    def hb_index(self):
        return self._hb_index

    @hb_index.setter
    def hb_index(self, hb_index):
        self._hb_index = hb_index
        if self._lb_index is not None:
            self.calc()

    @property
    def slice(self):
        return slice(self.lb_index, self.hb_index)

    def calc(self):
        """ Calculates the stats for the peak. """
        self.x = self._parent.result.index[self.slice].to_numpy()
        self.y = self._parent.result.iloc[self.slice].to_numpy()
        self.y_norm = self.y/np.trapz(x=self.x, y=self.y)
        self.lb_value = self.y[0]
        self.hb_value = self.y[-1]
        self.lb_loc = self.x[0]
        self.hb_loc = self.x[-1]

        self.max_index = np.argmax(self.y) + self.lb_index
        self.max = self.y[self.max_index-self.lb_index]
        self.max_loc = self.x[self.max_index-self.lb_index]

        self.mean = np.trapz(x=self.x, y=self.x*self.y_norm)
        self.std = np.sqrt(np.trapz(x=self.x, y=self.y_norm*(self.x-self.mean)**2))
        self.skew = np.trapz(x=self.x, y=self.y_norm * (self.x - self.mean) ** 3) / self.std ** 3
        self.kurtosis = (np.trapz(x=self.x, y=self.y_norm * (self.x - self.mean) ** 4) / self.std ** 4) - 3
        self.fwhm = self.get_fw(x=self.x, y=self.y, height=0.5)
        self.asym = self.get_asym(x=self.x, y=self.y, height=0.1)

        self.area = np.trapz(x=self.x, y=self.y)

    def get_fw(self, x: np.ndarray, y: np.ndarray, height: float = 0.5) -> Union[float, int]:
        """ Calculates full width at a height. """
        lower, high = self.get_width_at(x, y, height)
        return abs(high-lower)

    def get_asym(self, x: np.ndarray, y: np.ndarray, height: float = 0.1) -> Union[float, int]:
        """ Calculates asymmetry factor at height. """
        lower, high = self.get_width_at(x, y, height)
        middle = x[np.argmax(y)]

        return (high-middle)/(middle-lower)

    @staticmethod
    def get_width_at(x: np.ndarray, y: np.ndarray, height: float = 0.5) -> tuple[float, float]:
        """ Determine full-width-x_max of a peaked set of points, x and y. """
        height_half_max = np.max(y) * height
        index_max = np.argmax(y)
        if index_max == 0 or index_max == len(x):  # peak max is at end.
            logger_analysis.info("Finding fwhm is not possible with a peak max at an bound.")
            return 0, 0

        x_low = np.interp(height_half_max, y[:index_max], x[:index_max])
        x_high = np.interp(height_half_max, np.flip(y[index_max:]), np.flip(x[index_max:]))

        if x_low == x[0]:
            logger_analysis.info("fwhm or asym is having to linear interpolate on the lower end.")
            slice_ = max(3, int(index_max/10))
            fit = np.polyfit(y[:slice_], x[:slice_], deg=1)
            p = np.poly1d(fit)
            x_low = p(height_half_max)

        if x_high == x[-1]:
            logger_analysis.info("fwhm or asym is having to linear interpolate on the lower end.")
            slice_ = max(3, int(index_max/10))
            fit = np.polyfit(y[-slice_:], x[-slice_:], deg=1)
            p = np.poly1d(fit)
            x_high = p(height_half_max)

        return x_low, x_high

    def plot(self, fig: go.Figure = None, auto_open: bool = True, auto_format: bool = True,
             y_label: str = None, title: str = None, **kwargs) -> go.Figure:
        """ Plot

        General plotting

        Parameters
        ----------
        fig: go.Figure
            plotly figure; will automatically create if not provided
        auto_open: bool
            create "temp.html" and auto_open in browser
        auto_format: bool
            apply built-in formatting
        y_label: str
            y_axis label (used for multiple y-axis)
        title: str
            title

        Returns
        -------
        fig: go.Figure
            plotly figure

        """
        if fig is None:
            fig = go.Figure()

        if "color" in kwargs:
            color = kwargs.pop("color")
        else:
            color = 'rgb(0,0,0)'

        # add main trace
        plot_kwargs = {
            "x": self.x,
            "y": self.y,
            "mode": 'lines',
            "connectgaps": True,
            "name": self._parent.result.name,
            "line": dict(color=color)
        }
        if y_label is not None:
            plot_kwargs["yaxis"] = y_label

        fig.add_trace(go.Scatter(**plot_kwargs))

        if auto_format:
            if title is not None:
                fig.update_layout(title=title)
            add_plot_format(fig, self._parent.result.index.name, str(self._parent.result.name))

        if auto_open:
            global fig_count
            fig.write_html(f'temp{fig_count}.html', auto_open=True)
            fig_count += 1

        return fig

    def plot_add_on(self, fig: go.Figure, group: str = None, y_label: str = None, **kwargs):
        """ Plot

        Plots several things related to a peak.

        Parameters
        ----------
        fig: go.Figure
            plotly figure
        group: str
            the legend table group that the plot_add_on will be associated with (need to make things appear/disappear together)
        y_label: str
            Label for the y-axis the data is associated with

        """
        self._plot_max(fig, group, y_label, **kwargs)
        # self._plot_bounds(fig, group, **kwargs)
        self._plot_shade(fig, group, y_label, **kwargs)

    def _plot_shade(self, fig: go.Figure, group: str = None, y_label: str = None, **kwargs):
        """ Plots the shaded area for the peak. """
        kwargs_ = {
            "width": 0
        }
        if kwargs:
            kwargs_ = {**kwargs_, **kwargs}

        kkwargs = {}
        if group:
            kkwargs["legendgroup"] = group
        if y_label:
            kkwargs["yaxis"] = y_label

        fig.add_trace(go.Scatter(
            x=self.x,
            y=self.y,
            mode="lines",
            fill='tozeroy',
            line=kwargs_,
            showlegend=False,
            **kkwargs
        ))

    def _plot_max(self, fig: go.Figure, group: str = None, y_label: str = None, **kwargs):
        """ Plots peak name at max. """
        kwargs_ = {
            "size": 1
        }
        if kwargs:
            kwargs_ = {**kwargs_, **kwargs}

        kkwargs = {}
        if group:
            kkwargs["legendgroup"] = group
        if y_label:
            kkwargs["yaxis"] = y_label

        fig.add_trace(go.Scatter(
            x=[self.max_loc],
            y=[self.max],
            mode="text",
            marker=kwargs_,
            text=[f"{self._parent.name}: {self.id_}"],
            textposition="top center",
            showlegend=False,
            **kkwargs
        ))

    def _plot_bounds(self, fig: go.Figure, group: str = None, height: float = 0.08, y_label: str = None, **kwargs):
        """ Adds bounds at the bottom of the plot_add_on for peak area. """
        kwargs_ = {
            "width": 5
        }
        if kwargs:
            kwargs_ = {**kwargs_, **kwargs}

        kkwargs = {}
        if group:
            kkwargs["legendgroup"] = group
        if y_label:
            kkwargs["yaxis"] = y_label

        bound_height = max(self._parent.result) * height
        # bounds
        fig.add_trace(go.Scatter(
            x=[self.lb_loc, self.lb_loc],
            y=[-bound_height / 2, bound_height / 2],
            mode="lines",
            line=kwargs_,
            showlegend=False,
            **kkwargs

        ))
        fig.add_trace(go.Scatter(
            x=[self.hb_loc, self.hb_loc],
            y=[-bound_height / 2, bound_height / 2],
            mode="lines",
            line=kwargs_,
            showlegend=False,
            **kkwargs
        ))
        fig.add_trace(go.Scatter(
            x=[self.lb_loc, self.hb_loc],
            y=[0, 0],
            mode="lines",
            line=kwargs_,
            showlegend=False,
            **kkwargs
        ))

    def stats(self, op_print: bool = True, op_headers: bool = True, window: int = 150, headers: dict = None):
        """ Prints stats out for peak. """
        text = ""
        if headers is None:
            headers = {  # attribute: print
                "id_": "id", "lb_loc": "low bound", "max_loc": "max", "hb_loc": "high bound", "area": "area"
            }

        # format
        width = int(window / len(headers))
        row_format = ("{:<" + str(width) + "}") * len(headers)

        # headers
        if op_headers:
            headers_ = [_length_limit(head, width) for head in headers.values()]
            text = row_format.format(*headers_) + "\n"
            text = text + "-" * window + "\n"

        # values
        entries = [_length_limit(str(sig_figs(getattr(self, k), 3)), width) for k in headers]
        entries[0] = f"{self._parent.name}: {entries[0]}"  # peak name: peak id
        text = text + row_format.format(*entries) + "\n"

        if op_print:
            print(text)

        return text



from typing import Union, Callable

import pandas as pd
import plotly.graph_objs as go

from chem_analysis.analysis.base_obj.calibration import Cal
from chem_analysis.analysis.base_obj.chromatogram import Chromatogram
from chem_analysis.analysis.base_obj.signal_ import Signal
from chem_analysis.analysis.sec.sec_signal import SECSignal
from chem_analysis.analysis.utils.plot_format import get_plot_color, add_plot_format
from chem_analysis.analysis.utils import fig_count


class SECChrom(Chromatogram):

    _signal = SECSignal

    def __init__(self, data: Union[pd.DataFrame, Signal, list[Signal]], cal: Union[Cal, Callable] = None):
        if not isinstance(cal, Cal):
            cal = Cal(cal)
        self.cal = cal

        super().__init__(data)

    def plot(self, fig: go.Figure = None, auto_open: bool = True, auto_format: bool = True,
             op_peaks: bool = True, op_cal: bool = True, **kwargs) -> go.Figure:
        if fig is None:
            fig = go.Figure()

        colors = get_plot_color(self.num_signals)

        for i, (sig, color) in enumerate(zip(self, colors)):
            kwargs_ = {"color": color}
            if op_cal and i == 0:
                kwargs_["op_cal"] = True
            else:
                kwargs_["op_cal"] = False
            if kwargs:
                kwargs_ = {**kwargs_, **kwargs}
            fig = sig.plot_add_on(fig, auto_open=False, auto_format=False, op_peaks=op_peaks, **kwargs_)

        if auto_format:
            add_plot_format(fig, self.x_label, "; ".join(self.y_labels))

        if auto_open:
            global fig_count
            fig.write_html(f'temp{fig_count}.html', auto_open=True)
            fig_count += 1

        return fig


def local_run():
    from scipy.stats import norm
    import numpy as np

    def cal_func(time: np.ndarray):
        return 10**(0.0167 * time ** 2 - 0.9225 * time + 14.087)

    cal = Cal(cal_func, lb=900, ub=319_000)

    nx = 1000
    ny = 3
    x = np.linspace(0, 25, nx)
    y = np.empty((ny, nx))
    for i in range(ny):
        rv = norm(loc=15, scale=0.6)
        rv2 = norm(loc=18, scale=0.6)
        y[i, :] = 5 * np.linspace(0, 1, nx) + np.random.random(nx) + 100 * rv.pdf(x) + 20 * rv2.pdf(x)
    df = pd.DataFrame(data=y.T, index=x)
    df.columns = ["RI", "UV", "LS"]
    df.index.names = ["time"]

    chro = SECChrom(data=df, cal=cal)
    chro.auto_peak_baseline(deg=1)
    chro.plot()
    chro.stats()
    print("done")


if __name__ == "__main__":
    local_run()








from dash import html, dcc, dash_table


tab_load_layout = html.Div([
    html.H1('Load data'),
    dcc.Upload(
        id="upload-data",
        children=html.Div([
            'Drag and Drop or ',
            html.A('Select Files')
        ]),
        style={
            'width': '50%',
            'height': '60px',
            'lineHeight': '60px',
            'borderWidth': '1px',
            'borderStyle': 'dashed',
            'borderRadius': '5px',
            'textAlign': 'center',
            'margin': '10px'
        },
        multiple=False,  # Allow multiple files to be uploaded
    ),
    html.Div(id='output-data-upload'),

    dcc.RadioItems(id="ri_time"),

    dash_table.DataTable(
        id='data_table',
        style_table={'overflowX': 'auto'},
        style_cell={
            'height': 'auto',
            # all three widths are needed
            'minWidth': '60px', 'width': '100px', 'maxWidth': '150px',
            'whiteSpace': 'normal'
        },
                         ),
])

#         columns=[{"id": "time", "name": "time"}, {"id": "RI", "name": "RI"}],


from typing import Union

import numpy as np


def rolling_value(x: np.ndarray, peak_index: int = None, peak_value: Union[float, int] = None,
                  sensitivity: float = 0.03, cut_off: Union[None, float, int] = 0) -> tuple[int, int]:
    """

    Parameters
    ----------
    x: np.ndarray
        1D data to find bounds
    peak_index: int
        index of peak max
    peak_value: int, float
        value of peak max
    sensitivity: float
        How much it can go up before triggering a bound detection, fraction of max height
    cut_off: Union[None, float, int]
        When to stop if never goes to zero, fraction of max height

    Returns
    -------
    lb: int
        index of lower bound
    up: int
        index of upper bound
    """
    if peak_index is None:
        raise ValueError("Peak index or peak value are required.")
    if peak_value:
        peak_index_ = np.where(x == peak_value)
        if len(peak_index_[0]) == 0:
            raise ValueError(f"Peak not found in data. (peak_value: {peak_value})")
        elif len(peak_index_[0]) >= 2:
            raise ValueError(f"Multiple peaks have this peak_value. (peak_value: {peak_value})")

        peak_index = peak_index_[0]
    else:
        peak_value = x[peak_index]

    # set up cutoffs
    if cut_off is not None:
        cut_off_value = peak_value * cut_off
    else:
        cut_off_value = -10e-50
    if cut_off is not None:
        sensitivity_value = peak_value * sensitivity
    else:
        sensitivity_value = 10e-50

    # lower bound
    if peak_index == 0:
        lb = peak_index
    elif peak_index < 5:
        lb = np.argmin(x[:peak_index])
    else:
        min_ = x[peak_index]
        min_index = peak_index
        for i, v in enumerate(np.flip(x[:peak_index])):
            if v < min_:
                min_ = v
                min_index = peak_index - i
                if min_ < cut_off_value:
                    break
            if v - x[min_index] > sensitivity_value:
                break

        lb = min_index-1

    # upper bound
    if peak_index == len(x):
        ub = peak_index
    elif len(x) - peak_index < 5:
        ub = np.argmin(x[peak_index:])
    else:
        min_ = x[peak_index]
        min_index = peak_index
        for i, v in enumerate(x[peak_index:]):
            if v < min_:
                min_ = v
                min_index = peak_index + i
                if min_ < cut_off_value:
                    break
            if v - x[min_index] > sensitivity_value:
                break

        ub = min_index

    return lb, ub


bound_detection_methods = {
    "rolling_value": rolling_value
}



from dash.dependencies import Input, Output

from chem_analysis.server import app


# Tab 2 callback
@app.callback(Output('page-2-content', 'children'),
              [Input('page-2-radios', 'value')])
def page_2_radios(value):
    return 'You have selected "{}"'.format(value)



from typing import Callable

import numpy as np
import scipy.optimize
import plotly.graph_objs as go


class Cal:
    def __init__(self, cal: Callable, lb: float = None, ub: float = None, name: str = None):
        self.name = name
        self.cal = cal
        self.lb = lb
        self.ub = ub

        self.lb_loc = None
        self.ub_loc = None

        self.calc()

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self.cal(x)

    def calc(self):
        if self.ub:
            def _ub(x: np.ndarray) -> np.ndarray:
                return self(x) - self.ub

            result_ub = scipy.optimize.root_scalar(_ub, x0=0.1, x1=0.2)
            self.ub_loc = result_ub.root

        if self.lb:
            def _lb(x: np.ndarray) -> np.ndarray:
                return self(x) - self.lb

            if self.ub_loc:
                x0 = self.ub_loc
            else:
                x0 = 0.1

            result_lb = scipy.optimize.root_scalar(_lb, x0=x0, x1=x0 + 0.1)
            self.lb_loc = result_lb.root




from typing import Union


def sig_figs(number: Union[float, int, None], sig_digit: int = 3) -> Union[int, float, None]:
    """ significant figures
    Given a number return a string rounded to the desired significant digits.
    Parameters
    ----------
    number: float, int
        number you want to reduce significant figures on
    sig_digit: int
        significant digits
    Returns
    -------
    number: int, float
    """
    if isinstance(number, float):
        return float('{:.{p}g}'.format(number, p=sig_digit))
    elif isinstance(number, int):
        return int('{:.{p}g}'.format(number, p=sig_digit))
    elif number is None:
        return None
    else:
        raise TypeError(f"'sig_figs' only accepts int or float. Given: {number} (type: {type(number)}")



from dash.dependencies import Input, Output

from chem_analysis.server import app


# Tab 2 callback
@app.callback(Output('page-3-content', 'children'),
              [Input('page-3-radios', 'value')])
def page_2_radios(value):
    return 'You have selected "{}"'.format(value)
